from .address import Address
from .telegram import Telegram,TelegramType
from .dpt import DPT_Binary
from .exception import CouldNotParseTelegram
from .device import Device

class Outlet(Device):

    def __init__(self, xknx, name, config):
        self.group_address = Address(config["group_address"])
        Device.__init__(self, xknx, name)
        self.state = False


    def has_group_address(self, group_address):
        return self.group_address == group_address


    def set_internal_state(self, state):
        if state != self.state:
            self.state = state
            self.after_update_callback(self)


    def send(self, payload):
        telegram = Telegram()
        telegram.group_address=self.group_address
        telegram.payload = payload
        self.xknx.telegrams.put(telegram)

    def set_on(self):
        self.send( DPT_Binary(1) )
        self.set_internal_state(True)


    def set_off(self):
        self.send( DPT_Binary(0) )
        self.set_internal_state(False)


    def do(self,action):
        if(action=="on"):
            self.set_on()
        elif(action=="off"):
            self.set_off()
        else:
            print("{0}: Could not understand action {1}".format(self.get_name(), action))


    def request_state(self):
        telegram = Telegram(self.group_address, TelegramType.GROUP_READ)
        self.xknx.telegrams.put(telegram)

    def process(self,telegram):

        if not isinstance( telegram.payload, DPT_Binary ):
            raise CouldNotParseTelegram()

        if telegram.payload.value == 0 :
            self.set_internal_state(False)
        elif telegram.payload.value == 1 :
            self.set_internal_state(True)
        else:
            raise CouldNotParseTelegram()


    def __str__(self):
        return "<Outlet group_address={0}, name={1} internal_state={2}>".format(self.group_address,self.name,self.internal_state)




import threading
import time

from .telegram import TelegramDirection
from .devices import CouldNotResolveAddress
from .multicast import Multicast

class TelegramProcessor(threading.Thread):

    def __init__(self, xknx, telegram_received_callback = None):
        self.xknx = xknx
        self.telegram_received_callback = telegram_received_callback
        threading.Thread.__init__(self)


    def run(self):
        while True:
            telegram = self.xknx.telegrams.get()
            self.process_telegram(telegram)
            self.xknx.telegrams.task_done()

            if telegram.direction == TelegramDirection.OUTGOING:
                # limit rate to knx bus to 20 per second
                time.sleep(1/20)


    def process_telegram(self,telegram):
        if telegram.direction == TelegramDirection.INCOMING:
            self.process_telegram_incoming(telegram)
        elif telegram.direction == TelegramDirection.OUTGOING:
            self.process_telegram_outgoing(telegram)
            pass


    def process_telegram_outgoing(self,telegram):
        multicast = Multicast(self.xknx)
        multicast.send(telegram)


    def process_telegram_incoming(self,telegram):
        try:
            device = self.xknx.devices.device_by_group_address(telegram.group_address)
            device.process(telegram)
            if ( self.telegram_received_callback ):
                self.telegram_received_callback(self.xknx, device, telegram)
        except CouldNotResolveAddress as c:
            print(c)


    @staticmethod
    def start_thread(xknx, telegram_received_callback = None):
        t = TelegramProcessor(xknx,telegram_received_callback)
        t.setDaemon(True)
        t.start()



from .binaryinput import BinaryInput,BinaryInputState
from .device import Device
import yaml
import time
from enum import Enum
from .address import Address
from .devices import Devices

class SwitchTime(Enum):
    SHORT = 1
    LONG = 2

class Action():
    def __init__(self, xknx, config):
        self.xknx = xknx
        self.hook = config["hook"]
        self.target = config["target"]
        self.method = config["method"]
        if( "switch_time" in config ):
            if(config["switch_time"]=="long"):
                self.switch_time = SwitchTime.LONG
            elif(config["switch_time"]=="short"):
                self.switch_time = SwitchTime.SHORT

    def test_switch_time(self,switch_time):
        if not hasattr(self,"switch_time"):
            # no specific switch_time -> always true
            return True

        return (switch_time==self.switch_time)


    def test(self,state,switch_time):
        if((state==BinaryInputState.ON) and (self.hook == "on") and self.test_switch_time(switch_time)):
            return True

        if((state==BinaryInputState.OFF) and (self.hook == "off") and self.test_switch_time(switch_time)):
            return True
        return False

    def execute(self):
        self.xknx.devices.device_by_name(self.target).do(self.method)

    def __str__(self):
        return "<Action hook={0} target={1} method={2}>".format(self.hook,self.target,self.method)

class Switch(BinaryInput):
    def __init__(self, xknx, name, config):
        group_address = Address(config["group_address"])
        BinaryInput.__init__(self, xknx, name, group_address)
        self.group_address = group_address
        self.last_set = time.time();

        self.actions = []
        if "actions" in config:
            for action in config["actions"]:
                action = Action(xknx, action)
                self.actions.append(action)

    def get_switch_time(self):
        new_set_time = time.time()
        time_diff = new_set_time - self.last_set
        self.last_set = new_set_time
        if time_diff<0.2:
            return SwitchTime.SHORT
        return SwitchTime.LONG

    def process(self,telegram):
        BinaryInput.process(self,telegram)
        switch_time = self.get_switch_time()

        for action in self.actions:
            if(action.test(self.state,switch_time)):
                action.execute()

    def __str__(self):
        return "<Switch group_address={0}, name={1}>".format(self.group_address,self.name)




""" Implementation of Basic KNX datatypes """

""" See: http://www.knx.org/fileadmin/template/documents/downloads_support_menu/KNX_tutor_seminar_page/Advanced_documentation/05_Interworking_E1209.pdf as reference """

from enum import Enum
import time

class ConversionError(Exception):
    def __init__(self, i):
        self.i = i
    def __str__(self):
        return "<ConversionError input='{0}'>".format(self.i)


class DPT_Base:

    @staticmethod
    def test_bytesarray( raw,length ):
        if type(raw) is not tuple \
                or len(raw) != length \
                or any(not isinstance(byte,int) for byte in raw) \
                or any(byte < 0 for byte in raw) \
                or any(byte > 255 for byte in raw):
            raise ConversionError(raw)


class DPT_Binary(DPT_Base):
    """ The DPT_Binary is a base class for all datatypes encoded
    directly into the first Byte of the payload """

    # APCI (application layer control information)  
    APCI_BITMASK = 0x3F
    APCI_MAX_VALUE = APCI_BITMASK 

    def __init__( self, value ):
        if not isinstance(value,int):
            raise TypeError()
        if value > DPT_Binary.APCI_BITMASK:
            raise ConversionError(value)

        self.value = value

    def __eq__(self, other):
        return DPT_Comparator.compare(self,other)

    def __str__(self):
        return "<DPT_Binary value={0}>".format(self.value)



class DPT_Array(DPT_Base):
    """ The DPT_Array is a base class for all datatypes appended
    to the KNX telegram """

    def __init__(self, value ):
        if isinstance(value, int):
            self.value = (value,)
        elif isinstance(value, (list,bytes)):
            self.value = tuple(value,)
        elif isinstance(value, tuple):
            self.value = value
        else:
            print(value,type(value))
            raise TypeError()

    def __eq__(self, other):
        return DPT_Comparator.compare(self,other)

    def __str__(self):
        return "<DPT_Array value=[{0}]>".format( ','.join(hex(b) for b in self.value))



class DPT_Comparator():
    """ Helper class to compare different types of DPT objects"""

    @staticmethod
    def compare(a,b):
        if a is None and b is None:
            return True

        elif a is None:
            if isinstance(b,DPT_Binary):
                return b.value == 0
            elif isinstance(b,DPT_Array):
                return len(b.value) == 0

        elif b is None:
            if isinstance(a,DPT_Binary):
                return a.value == 0
            elif isinstance(a,DPT_Array):
                return len(a.value) == 0

        elif isinstance(a,DPT_Array) and isinstance(b,DPT_Array):
            return a.value == b.value

        elif isinstance(a,DPT_Binary) and isinstance(b,DPT_Binary):
            return a.value == b.value

        elif isinstance(a,DPT_Binary) and isinstance(b,DPT_Array):
            return a.value == 0 and len(b.value) == 0

        elif isinstance(a,DPT_Array) and isinstance(b,DPT_Binary):
            return len(a.value) == 0 and b.value == 0

        raise TypeError()




import time
from .colors import Colors
from .address import Address,AddressType
from .telegram import Telegram,TelegramType
from .knxip_enum import KNXIPServiceType,CEMIMessageCode,APCICommand,CEMIFlags
from .dpt import DPT_Binary,DPT_Array

# # See: http://www.knx.org/fileadmin/template/documents/downloads_support_menu/KNX_tutor_seminar_page/tutor_documentation/08_IP%20Communication_E0510a.pdf

class CouldNotParseKNXIP(Exception):
    def __init__(self, description = ""):
        self.description = description
    def __str__(self):
        return "<CouldNotParseKNXIP description='{0}'>".format(self.description)

class ConversionException(Exception):
    def __init__(self, description = ""):
        self.description = description
    def __str__(self):
        return "<ConversionException description='{0}'>".format(self.description)

class ConnectionHeader():

    HEADERLENGTH = 0x06
    PROTOCOLVERSION = 0x10

    def __init__(self, data = None):
        self.headerLength = ConnectionHeader.HEADERLENGTH
        self.protocolVersion = ConnectionHeader.PROTOCOLVERSION
        self.serviceTypeIdent = KNXIPServiceType.ROUTING_INDICATION
        self.b4Reserve = 0
        self.totalLength = 0 # to be set later


    def from_knx(self, data):
        if len(data) != 6:
            raise CouldNotParseKNXIP("wrong connection header length")
        if data[0] != ConnectionHeader.HEADERLENGTH:
            raise CouldNotParseKNXIP("wrong connection header length")
        if data[1] != ConnectionHeader.PROTOCOLVERSION:
            raise CouldNotParseKNXIP("wrong protocol version")

        self.headerLength = data[0]
        self.protocolVersion = data[1]
        self.serviceTypeIdent = KNXIPServiceType(data[2] * 256 + data[3])
        self.b4Reserve = data[4]
        self.totalLength = data[5]

    def set_length( self, cemi ):
        if not isinstance(cemi, CEMIFrame):
            raise TypeError()
        self.totalLength = ConnectionHeader.HEADERLENGTH + cemi.calculated_length()

    def to_knx(self):
        data = []

        data.append( self.headerLength)
        data.append( self.protocolVersion)
        data.append( ( self.serviceTypeIdent.value >> 8 ) & 255 )
        data.append( self.serviceTypeIdent.value & 255 )
        data.append( ( self.totalLength>>8 ) & 255 )
        data.append (self.totalLength & 255 )

        return data

    def __str__(self):
        return "<Connection HeaderLength={0}, ProtocolVersion={1}, " \
                "KNXIPServiceType={2}, Reserve={3}, TotalLength={4}>".format(
                self.headerLength, self.protocolVersion, self.serviceTypeIdent,
                self.b4Reserve, self.totalLength)



class CEMIFrame():
    """Representation of a CEMI Frame."""

    def __init__(self):
        """CEMIFrame __init__ object."""
        self.code = CEMIMessageCode.L_Data_REQ
        self.flags = 0
        self.cmd = APCICommand.GROUP_READ
        self.src_addr = Address()
        self.dst_addr = Address()
        self.mpdu_len = 0
        self.payload = None

    def set_hops(self,hops):
        # Resetting hops
        self.flags &= 0xFFFF ^ 0x0070
        # Setting new hops
        self.flags |= hops << 4

    def calculated_length(self):
        if self.payload is None:
            return 11
        elif isinstance( self.payload, DPT_Binary ):
            return 11
        elif isinstance(self.payload, DPT_Array ):
            return 11 + len(self.payload.value)
        else:
            raise TypeError()

    def from_knx(self, cemi):

        """Create a new CEMIFrame initialized from the given CEMI data."""
        # TODO: check that length matches
        self.code = CEMIMessageCode(cemi[0])
        offset = cemi[1]

        self.flags = cemi[2] * 256 + cemi[3]

        self.src_addr = Address((cemi[4 + offset], cemi[5 + offset]), AddressType.PHYSICAL)

        dst_addr_type = \
            AddressType.GROUP \
            if self.flags & CEMIFlags.DESTINATION_GROUP_ADDRESS \
            else AddressType.PHYSICAL
        self.dst_addr = Address((cemi[6 + offset], cemi[7 + offset]), dst_addr_type)

        self.mpdu_len = cemi[8 + offset]

        # TPCI (transport layer control information)   -> First 14 bit
        # APCI (application layer control information) -> Last  10 bit 

        tpci_apci = cemi[9 + offset] * 256 + cemi[10 + offset]

        self.cmd = APCICommand(tpci_apci & 0xFFC0)

        apdu = cemi[10 + offset:]
        if len(apdu) != self.mpdu_len:
            raise CouldNotParseKNXIP(
                "APDU LEN should be {} but is {}".format(
                    self.mpdu_len, len(apdu)))

        if len(apdu) == 1:
            apci = tpci_apci & DPT_Binary.APCI_BITMASK
            self.payload = DPT_Binary( apci )
        else:
            self.payload = DPT_Array( cemi[11 + offset:] )

    def to_knx(self):
        """Convert the CEMI frame object to its byte representation."""

        data = []

        data.append(self.code.value)
        data.append(0x00)
        data.append((self.flags >> 8 )& 255)
        data.append(self.flags & 255)
        data.append(self.src_addr.byte1()& 255)
        data.append(self.src_addr.byte2()& 255)
        data.append(self.dst_addr.byte1()& 255)
        data.append(self.dst_addr.byte2()& 255)		

        def encode_cmd_and_payload( cmd, encoded_payload = 0, appended_payload = []):
            data = [
                1 + len( appended_payload ),
                (cmd.value >> 8) & 0xff,
                (cmd.value & 0xff) | ( encoded_payload & DPT_Binary.APCI_BITMASK ) ]
            data.extend(appended_payload)
            return data

        if self.payload is None:
            data.extend( encode_cmd_and_payload( self.cmd ) )
        elif isinstance( self.payload, DPT_Binary ):
            data.extend( encode_cmd_and_payload( self.cmd, encoded_payload = self.payload.value ) )
        elif isinstance( self.payload, DPT_Array ):
            data.extend( encode_cmd_and_payload( self.cmd, appended_payload = self.payload.value ) )
        else:
            raise TypeError()

        return data


    def __str__(self):
            return "<CEMIFrame SourceAddress={0}, DestinationAddress={1}, " \
                   "Flags={2:16b} Command={3}, payload={4}>".format( self.src_addr, self.dst_addr,
                   self.flags, self.cmd, self.payload)


class KNXIPFrame:
    """Abstraction for KNX IP Frames"""


    def __init__(self):
        """Initialize object."""

        self.cemi = CEMIFrame()
        self.header = ConnectionHeader()

        self.payload = None

    @property
    def sender(self):
        """Return source address"""
        return self.cemi.src_addr

    @sender.setter
    def sender(self, sender):
        self.cemi.src_addr = Address(sender, AddressType.PHYSICAL)

    @property
    def group_address(self):
        """Return destination address"""
        return self.cemi.dst_addr

    @group_address.setter
    def group_address(self, group_address):
        self.cemi.dst_addr = Address(group_address, AddressType.GROUP)

    @property
    def telegram(self):
        telegram = Telegram()
        telegram.payload = self.cemi.payload
        telegram.group_address = self.group_address

        if self.cemi.cmd == APCICommand.GROUP_WRITE:
            telegram.type = TelegramType.GROUP_WRITE
        elif self.cemi.cmd == APCICommand.GROUP_READ:
            telegram.type = TelegramType.GROUP_READ
        elif self.cemi.cmd == APCICommand.GROUP_RESPONSE:
            telegram.type = TelegramType.GROUP_RESPONSE
        else:
            raise ConversionException("Telegram not implemented for {0}".format(self.cemi.cmd))

        # TODO: Set telegram.direction [additional flag within KNXIP]
        return telegram

    @telegram.setter
    def telegram(self,telegram):
        self.cemi.dst_addr = telegram.group_address
        self.cemi.payload = telegram.payload

        # TODO: Move to separate function
        self.cemi.code = CEMIMessageCode.L_DATA_IND
        self.cemi.flags = ( CEMIFlags.FRAME_TYPE_STANDARD | CEMIFlags.DO_NOT_REPEAT |
                    CEMIFlags.BROADCAST | CEMIFlags.PRIORITY_LOW |
                    CEMIFlags.NO_ACK_REQUESTED | CEMIFlags.CONFIRM_NO_ERROR |
                    CEMIFlags.DESTINATION_GROUP_ADDRESS | CEMIFlags.HOP_COUNT_1ST)

        # TODO: use telegram.direction
        if telegram.type == TelegramType.GROUP_READ:
            self.cemi.cmd = APCICommand.GROUP_READ
        elif telegram.type == TelegramType.GROUP_WRITE:
            self.cemi.cmd = APCICommand.GROUP_WRITE
        elif telegram.type == TelegramType.GROUP_RESPONSE:
            self.cemi.cmd = APCICommand.GROUP_RESPONSE
        else:
            raise TypeError()


    def from_knx(self, data):

        self.header.from_knx(data[0:6])
        self.cemi.from_knx(data[6:])

    def __str__(self):
        return "<KNXIPFrame {0}\n{1}>".format(self.header, self.cemi)


    def normalize(self):
        self.header.set_length( self.cemi )

    def to_knx(self):

        data = []
        data.extend(self.header.to_knx())
        data.extend(self.cemi.to_knx())
        return data;



import socket
import struct
import threading
import time

from .telegram import Telegram,TelegramDirection
from .knxip import KNXIPFrame
from .address import Address
from .xknx import XKNX
from .devices import CouldNotResolveAddress

class Multicast:
    MCAST_GRP = '224.0.23.12'
    MCAST_PORT = 3671

    def __init__(self, xknx):
        self.xknx = xknx

    def send(self, telegram):

        knxipframe = KNXIPFrame()
        knxipframe.telegram = telegram
        knxipframe.sender = self.xknx.globals.own_address
        knxipframe.normalize()

        self._send_knxipframe(knxipframe)

    def _send_knxipframe( self, knxipframe):
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)

        if self.xknx.globals.own_ip is not None:
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton(self.xknx.globals.own_ip))

        knxipframe.normalize()
        sock.sendto(bytes(knxipframe.to_knx()), (self.MCAST_GRP, self.MCAST_PORT))

    def start_daemon(self):
        print("Starting daemon...")
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.bind(("", self.MCAST_PORT))

        if self.xknx.globals.own_ip is not None:
            sock.setsockopt(socket.IPPROTO_IP,
                                 socket.IP_ADD_MEMBERSHIP,
                                 socket.inet_aton(self.MCAST_GRP) +
                                 socket.inet_aton(self.xknx.globals.own_ip))
        else:
            mreq = struct.pack("4sl", socket.inet_aton(self.MCAST_GRP), socket.INADDR_ANY)
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)

        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 0)

        return sock

    def recv(self,sock):
        while True:
            raw = sock.recv(10240)
            if raw:

                if len(raw) < 17:
                    print("WARNING: KNXIPFrame has size {0} too small, ignoring".format(len(raw)))
                    continue

                knxipframe = KNXIPFrame()
                knxipframe.from_knx(raw)

                if knxipframe.sender == self.xknx.globals.own_address:
                    # Ignoring own KNXIPFrame
                    pass

                else:
                    telegram = knxipframe.telegram
                    # TODO: This should be inside knxipframe
                    telegram.direction = TelegramDirection.INCOMING

                    self.xknx.telegrams.put(telegram)

class MulticastDaemon(threading.Thread):
    def __init__(self, xknx):
        self.xknx = xknx
        threading.Thread.__init__(self)


    def run(self):
        multicast = Multicast(self.xknx)
        sock = multicast.start_daemon()
        multicast.recv(sock)


    @staticmethod
    def start_thread(xknx):
        t = MulticastDaemon(xknx)
        t.setDaemon(True)
        t.start()




from enum import Enum
import time

class PositionType(Enum):
    UNKNOWN = 1
    CALCULATED = 2
    CONFIRMED = 3

class TravelStatus(Enum):
    DIRECTION_UP = 1
    DIRECTION_DOWN = 2
    STOPPED = 3


class TravelCalculator:


    def  __init__(self, travel_time_down, travel_time_up):
        self.position_type = PositionType.UNKNOWN
        self.last_known_position = 0

        self.travel_time_down = travel_time_down
        self.travel_time_up  = travel_time_up

        self.travel_to_position = 0
        self.travel_started_time = 0
        self.travel_direction = TravelStatus.STOPPED
        #TODO: Move to DPT Types class issue #10
        self.minimum_position_down = 256 # excluding
        self.maximum_position_up = 0

    def set_position(self, position):
        self.last_known_position = position
        self.travel_to_position = position
        self.position_type = PositionType.CONFIRMED

    def stop(self):
        self.last_known_position = self.current_position()
        self.travel_to_position = self.last_known_position
        self.position_type = PositionType.CALCULATED
        self.travel_direction = TravelStatus.STOPPED

    def start_travel(self, travel_to_position ):
        self.stop()

        self.travel_started_time = self.current_time()
        self.travel_to_position = travel_to_position
        self.position_type = PositionType.CALCULATED

        if travel_to_position < self.last_known_position:
            self.travel_direction = TravelStatus.DIRECTION_UP
        else:
            self.travel_direction = TravelStatus.DIRECTION_DOWN

    def start_travel_up(self):
        self.start_travel(self.maximum_position_up)

    def start_travel_down(self):
        self.start_travel(self.minimum_position_down)

    def current_position(self):
        if ( self.position_type == PositionType.CALCULATED ):
            return self._calculate_position()
        return self.last_known_position

    def is_travelling(self):
        return self.current_position() !=  self.travel_to_position

    def position_reached(self):
        return self.current_position() ==  self.travel_to_position

    def is_open(self):
        return self.current_position() == self.maximum_position_up

    def is_closed(self):
        return self.current_position() == self.minimum_position_down

    def _calculate_position(self):
        relative_position = self.travel_to_position - self.last_known_position

        def position_reached_or_exceeded(relative_position):
            if relative_position <= 0 and self.travel_direction == TravelStatus.DIRECTION_DOWN :
                return True
            if relative_position >= 0 and self.travel_direction == TravelStatus.DIRECTION_UP :
                return True
            return False

        if position_reached_or_exceeded(relative_position):
             return self.travel_to_position

        travel_time = self._calculate_travel_time( relative_position )
        if self.current_time() > self.travel_started_time + travel_time:
            return self.travel_to_position
        progress = (self.current_time()-self.travel_started_time)/travel_time
        position = self.last_known_position + relative_position * progress
        return int(position)

    def _calculate_travel_time(self, relative_position):
        travel_direction = \
                    TravelStatus.DIRECTION_UP \
                    if relative_position < 0 else \
                    TravelStatus.DIRECTION_DOWN
        travel_time_full = \
                    self.travel_time_up \
                    if travel_direction == TravelStatus.DIRECTION_UP else \
                    self.travel_time_down
        return travel_time_full * abs(relative_position) / (self.minimum_position_down-self.maximum_position_up)

    def current_time(self):
        # time_set_from_outside is  used within unit tests
        if hasattr(self, 'time_set_from_outside'):
            return self.time_set_from_outside
        return time.time()






""" Implementation of Basic KNX Time """

""" See: http://www.knx.org/fileadmin/template/documents/downloads_support_menu/KNX_tutor_seminar_page/Advanced_documentation/05_Interworking_E1209.pdf as reference """

from enum import Enum
import time

from .dpt import DPT_Base,ConversionError

class DPT_Weekday(Enum):
    MONDAY    = 1
    TUESDAY   = 2
    WEDNESDAY = 3
    THURSDAY  = 4
    FRIDAY    = 5
    SATURDAY  = 6
    SUNDAY    = 7
    NONE      = 0

class DPT_Time(DPT_Base):
    """ Abstraction for KNX 3 Octet Time """
    """ DPT 10.001 """

    @classmethod
    def from_knx(cls,raw):
        """Convert a 3 byte KNX date to a time dict"""

        cls.test_bytesarray(raw,3)

        day     = ( raw[0] & 0xE0 ) >> 5
        hours   = raw[0] & 0x1F
        minutes = raw[1] & 0x3F
        seconds = raw[2] & 0x3F

        if not DPT_Time.test_range(day,hours,minutes,seconds):
            raise ConversionError(raw)

        return {'day':DPT_Weekday(day),'hours':hours,
            'minutes':minutes,'seconds':seconds}

    @classmethod
    def to_knx(cls,values):
        """Convert time tuple to KNX time
        @param value: dict with following elements: (day,hours,minutes,seconds)
        """

        if not isinstance(values, dict):
            raise ConversionError(values)

        day     = values.get('day',DPT_Weekday.NONE).value
        hours   = values.get('hours',0)
        minutes = values.get('minutes',0)
        seconds = values.get('seconds',0)

        if not DPT_Time.test_range(day,hours,minutes,seconds):
            raise ConversionError(value)

        return day << 5 | hours, minutes, seconds

    @classmethod
    def current_time(cls):
        t = time.localtime()

        day     = t.tm_wday + 1
        hours   = t.tm_hour
        minutes = t.tm_min
        seconds = t.tm_sec

        return {'day':DPT_Weekday(day),'hours':hours,
            'minutes':minutes,'seconds':seconds}

    @classmethod
    def current_time_as_knx(cls):
        return cls.to_knx( cls.current_time() )

    @staticmethod
    def test_range(day,hours,minutes,seconds):
        if day < 0 or day > 7:
            return False
        if hours < 0 or hours > 23:
            return False
        if minutes  < 0 or minutes > 59:
            return False
        if seconds  < 0 or seconds > 59:
            return False
        return True



import yaml
from .xknx import XKNX
from .address import Address
from .binaryinput import BinaryInput
from .device import Device
from .switch import Switch
from .thermostat import Thermostat
from .time import Time
from .dimmer import Dimmer
from .outlet import Outlet
from .shutter import Shutter
from .devices import Devices
from .address import Address,AddressType
import time


class Config:

    def __init__(self, xknx):
        self.xknx = xknx

    def read( self, file='xknx.yaml'):
        print("Reading {0}".format(file))
        with open(file, 'r') as f:
            doc = yaml.load(f)
            self.parse_general(doc)
            self.parse_groups(doc)

    def parse_general(self, doc):
        if "general" in doc:
            if "own_address" in doc["general"]:
                self.xknx.globals.own_address = Address(doc["general"]["own_address"],AddressType.PHYSICAL)
            if "own_ip" in doc["general"]:
                self.xknx.globals.own_ip = doc["general"]["own_ip"]

    def parse_groups(self, doc):
        for group in doc["groups"]:
            if group.startswith("dimmer"):
                for entry in doc["groups"][group]:
                    dimmer = Dimmer(self.xknx, entry, doc["groups"][group][entry])
                    self.xknx.devices.devices.append(dimmer)
            if group.startswith("outlet"):
                for entry in doc["groups"][group]:
                    outlet = Outlet(self.xknx, entry, doc["groups"][group][entry])
                    self.xknx.devices.devices.append(outlet)
            if group.startswith("switch"):
                for entry in doc["groups"][group]:
                    switch = Switch(self.xknx, entry, doc["groups"][group][entry])
                    self.xknx.devices.devices.append(switch)
            if group.startswith("shutter"):
                for entry in doc["groups"][group]:
                    shutter = Shutter(self.xknx, entry, doc["groups"][group][entry])
                    self.xknx.devices.devices.append(shutter)

            if group.startswith("thermostat"):
                for entry in doc["groups"][group]:
                    thermostat = Thermostat(self.xknx, entry, doc["groups"][group][entry])
                    self.xknx.devices.devices.append(thermostat)
            if group.startswith("time"):
                for entry in doc["groups"][group]:
                    time = Time(self.xknx, entry, doc["groups"][group][entry])
                    self.xknx.devices.devices.append(time)



from .address import Address,AddressType

class Globals():

    def __init__(self):
        """ own_address is the own group address of the XKNX daemon """
        self.own_address = Address("15.15.250", AddressType.PHYSICAL)

        """ own_ip is the own ip address of the daemon. Needed if multiple 
        network interfaces are available to detect on which the multicast daemon 
        should listen """
        self.own_ip = None




from .device import Device
from .address import Address
from .dpt_time import DPT_Time
from .telegram import Telegram
from .dpt import DPT_Array
import time


class Time(Device):
    def __init__(self, xknx, name, config):
        Device.__init__(self, xknx, name)
        self.group_address = Address(config["group_address"])

    def has_group_address(self, group_address):
        return self.group_address == group_address

    def broadcast_time(self):
        telegram = Telegram()
        telegram.group_address=self.group_address
        telegram.payload = DPT_Array(DPT_Time.current_time_as_knx())
        self.xknx.telegrams.put(telegram)

    def request_state(self):
        self.broadcast_time()

    def __str__(self):
        return "<Time group_address={0}, name={1}>".format(self.group_address,self.name)



from .address import Address
from .dpt import DPT_Binary
from enum import Enum

class TelegramDirection(Enum):
    INCOMING = 1
    OUTGOING = 2

class TelegramType(Enum):
    GROUP_READ = 1
    GROUP_WRITE = 2
    GROUP_RESPONSE = 3

class Telegram:
    """ Abstraction for the business logic of KNX Telegrams

        This is a leightweight object for business logic,
        only containing group address and payload.

        The telegram direction marks if a telegram originates from
        XKNX abstraction or was sent from outside for internal
        processing.
    """

    def __init__(self, group_address = Address(),
            type = TelegramType.GROUP_WRITE,
            direction = TelegramDirection.OUTGOING, payload = None ):
        self.direction = direction
        self.type = type
        self.group_address = group_address
        self.payload = payload

    def __str__(self):
        return "<Telegram group_address={0}, payload={1} type={2} direction={3}>".format(
            self.group_address,
            self.payload,
            self.type,
            self.direction)

    # TODO: Add unit test
    def __eq__(self, other):
        return self.__dict__ == other.__dict__



from .device import Device
from .dpt import DPT_Binary
from .exception import CouldNotParseTelegram

from enum import Enum

class BinaryInputState(Enum):
    ON = 1
    OFF = 2

class BinaryInput(Device):
    def __init__(self, xknx, name, group_address):
        Device.__init__(self, xknx, name)
        self.group_address=group_address
        self.state=BinaryInputState.OFF

    def has_group_address(self, group_address):
        return self.group_address == group_address

    def set_internal_state(self, state):
        if state != self.state:
            self.state = state


    def process(self, telegram ):
        if not isinstance( telegram.payload, DPT_Binary ):
            raise CouldNotParseTelegram()

        if telegram.payload.value == 0:
            self.set_internal_state( BinaryInputState.OFF )
        elif telegram.payload.value == 1:
            self.set_internal_state( BinaryInputState.ON )
        else:
            raise CouldNotParseTelegram()


    def is_on(self):
        return self.state == BinaryInputState.ON


    def is_off(self):
        return self.state == BinaryInputState.OFF



from enum import Enum

class CouldNotParseAddress(Exception):
    def __init__(self, address = None):
        self.address = address

    def __str__(self):
        return "<CouldNotParseAddress address='{0}'>".format(self.address)

class AddressType(Enum):
    PHYSICAL = 1
    GROUP = 2

class AddressFormat(Enum):
    LEVEL2 = 1
    LEVEL3 = 2
    FREE = 3

class Address:

    def __init__(self, address = 0, address_type = None):
        self._set(address, address_type)

    def __eq__(self, other):
        if type(other) is not Address:
            raise TypeError()
        return self.raw == other.raw

    def __str__(self):
        return self._to_str()

    def _set(self, address, address_type):

        self.address_type = \
            address_type if address_type is not None \
            else self._detect_address_type(address)

        if address is None:
            self.raw=0
            self.address_format = AddressFormat.LEVEL3

        elif type(address) is Address:
            self.raw = address.raw
            self.address_format = address.address_format

        elif type(address) is str:
            if self.address_type == AddressType.PHYSICAL:
                self._set_str_physical(address)
            else:
                self._set_str_group(address)

        elif type(address) is int:
            self._set_int(address)
            self.address_format = AddressFormat.FREE

        elif type(address) is tuple:
            self._set_tuple(address)
            self.address_format = AddressFormat.FREE

        else:
            raise TypeError()

    def byte1(self):
        return ( self.raw >> 8 ) & 255

    def byte2(self):
        return self.raw & 255

    def is_set(self):
        return self.raw != 0

    ##################################################
    def _set_str_physical( self, address ):
        parts = address.split(".")
        if any(not part.isdigit() for part in parts):
            raise CouldNotParseAddress(address)
        if len(parts) != 3:
            raise CouldNotParseAddress(address)
        main = int(parts[0])
        middle = int(parts[1])
        sub = int(parts[2])
        if main > 15:
            raise CouldNotParseAddress(address)
        if middle > 15:
            raise CouldNotParseAddress(address)
        if sub > 255:
            raise CouldNotParseAddress(address)
        self.raw = (main<<12) +  (middle<<8) + sub
        self.address_format = AddressFormat.LEVEL3


    def _set_str_group( self, address ):
        parts = address.split("/")

        if any(not part.isdigit() for part in parts):
            raise CouldNotParseAddress(address)
        if len(parts) == 1:
            self._set_int( int ( parts[0] ) )
        elif len(parts) == 2:
            self._set_str_group_level2( parts )
        elif len(parts) == 3:
            self._set_str_group_level3( parts )
        else:
            raise CouldNotParseAddress(address)

    def _set_str_group_level2( self, parts ):
        main = int(parts[0])
        sub = int(parts[1])
        if main > 31:
            raise CouldNotParseAddress(parts)
        if sub > 2047:
            raise CouldNotParseAddress(parts)
        self.raw = (main<<11) + sub
        self.address_format = AddressFormat.LEVEL2

    def _set_str_group_level3( self, parts ):
        main = int(parts[0])
        middle = int(parts[1])
        sub = int(parts[2])
        if main > 31:
            raise CouldNotParseAddress(parts)
        if middle > 7:
            raise CouldNotParseAddress(parts)
        if sub > 255:
            raise CouldNotParseAddress(parts)
        self.raw = (main<<11) +  (middle<<8) + sub
        self.address_format = AddressFormat.LEVEL3

    def _set_tuple(self, address):
        if len(address) != 2 \
                or any(not isinstance(byte,int) for byte in address) \
                or any(byte < 0 for byte in address) \
                or any(byte > 255 for byte in address):
            raise CouldNotParseAddress(address)

        self._set_int(address[0] * 256 + address[1])

    def _set_int(self, raw):
        if type(raw) is not int:
            raise CouldNotParseAddress(raw)
        if ( raw > 65535 ):
            raise CouldNotParseAddress(raw)
        self.raw = raw
        self.address_format = AddressFormat.FREE

    def _to_str(self):
        if self.address_type == AddressType.PHYSICAL:
            return self._to_str_physical()
        elif self.address_format == AddressFormat.FREE:
            return self._to_str_free()
        elif self.address_format == AddressFormat.LEVEL2:
            return self._to_str_level2()
        elif self.address_format == AddressFormat.LEVEL3:
            return self._to_str_level3()
        else:
            raise TypeError()

    def _to_str_free(self):
        return '{0}'.format(
            (self.raw & 65535) )

    def _to_str_level2(self):
        return '{0}/{1}'.format(
            ((self.raw >> 11 ) & 15),
            (self.raw & 4095) )

    def _to_str_level3(self):
        return '{0}/{1}/{2}'.format(
            ((self.raw >> 11 ) & 31),
            ((self.raw >> 8) & 7),
            (self.raw & 255) )

    def _to_str_physical(self):
        return '{0}.{1}.{2}'.format(
            ((self.raw >> 12 ) & 15),
            ((self.raw >> 8) & 15),
            (self.raw & 255) )

    @staticmethod
    def _detect_address_type(address):
        # Physical addresses have either be specified explicitely or
        # in the correct notation. As default an address is a group address.
        if type(address) is str and "." in address:
            return AddressType.PHYSICAL
        elif type(address) is Address:
            return address.address_type
        else:
            return AddressType.GROUP



import queue
from .devices import Devices
from .globals import Globals
import signal

class XKNX:

    START_TELEGRAM_PROCESSOR = 0x01
    START_MULITCAST_DAEMON   = 0x02
    START_STATE_UPDATER      = 0x04

    START_DEFAULT = START_TELEGRAM_PROCESSOR | \
                    START_MULITCAST_DAEMON | \
                    START_STATE_UPDATER

    def __init__(self):
        self.globals = Globals()
        self.devices = Devices()
        self.telegrams = queue.Queue()


    def start(self,
            daemon_mode = False,
            start = START_DEFAULT,
            telegram_received_callback = None):
        from .multicast import MulticastDaemon
        from .telegram_processor import  TelegramProcessor
        from .stateupdater import StateUpdater

        if start & XKNX.START_TELEGRAM_PROCESSOR:
            TelegramProcessor.start_thread(self, telegram_received_callback)
        if start & XKNX.START_MULITCAST_DAEMON:
            MulticastDaemon.start_thread(self)
        if start & XKNX.START_STATE_UPDATER:
            StateUpdater.start_thread(self, 10)

        if daemon_mode:
            self.loop_until_sigint()


    def join(self):
        """ Wait until all telegrams were processed """
        self.telegrams.join()


    def stop(self):
        self.join()
        print("Process stopped")


    def loop_until_sigint(self):
        def sigint_handler(signal, frame):
            self.stop()

        signal.signal(signal.SIGINT, sigint_handler)
        print('Press Ctrl+C')
        signal.pause()



from enum import Enum



class KNXIPServiceType(Enum):

    SEARCH_REQUEST = 0x0201
    SEARCH_RESPONSE = 0x0202
    DESCRIPTION_REQUEST = 0x0203
    DESCRIPTION_RESPONSE = 0x0204
    CONNECT_REQUEST = 0x0205
    CONNECT_RESPONSE = 0x0206
    CONNECTIONSTATE_REQUEST = 0x0207
    CONNECTIONSTATE_RESPONSE = 0x0208
    DISCONNECT_REQUEST = 0x0209
    DISCONNECT_RESPONSE = 0x020a
    DEVICE_CONFIGURATION_REQUEST = 0x0310
    DEVICE_CONFIGURATION_ACK = 0x0111
    TUNNELING_REQUEST = 0x0420
    TUNNELLING_ACK = 0x0421
    ROUTING_INDICATION = 0x0530
    ROUTING_LOST_MESSAGE = 0x0531
    UNKNOWN = 0x0000


class CEMIMessageCode(Enum):

    # FROM NETWORK LAYER TO DATA LINK LAYER
    L_RAW_REQ       =   0x10
    L_Data_REQ      =   0x11 # Data Service. Primitive used for transmitting a data frame
    L_POLL_DATA_REQ =   0x13 # Poll Data Service

    # FROM DATA LINK LAYER TO NETWORK LAYER
    L_POLL_DATA_CON =   0x25 # Poll Data Service
    L_DATA_IND      =   0x29 # Data Service. Primitive used for receiving a data frame
    L_BUSMON_IND    =   0x2B # Bus Monitor Service
    L_RAW_IND       =   0x2D
    L_DATA_CON      =   0x2E # Data Service. Primitive used for local confirmation that a frame was sent#
                             # (does not indicate a successful receive though)
    L_RAW_CON       =   0x2F


class APCICommand(Enum):

    GROUP_READ     = 0x00
    GROUP_WRITE    = 0x80
    GROUP_RESPONSE = 0x40



class CEMIFlags():

        # Bit 1/7
        FRAME_TYPE_EXTENDED = 0x0000
        FRAME_TYPE_STANDARD = 0x8000

        # Bit 1/6 - Reserved

        # Bit 1/5
        # Repeat in case of an error
        REPEAT        = 0x0000
        DO_NOT_REPEAT = 0x2000

        # Bit 1/4
        SYSTEM_BROADCAST = 0x0000
        BROADCAST        = 0x1000


        # Bit 1/3+2
        PRIORITY_SYSTEM    = 0x0000
        PRIORITY_NORMAL    = 0x0400
        PRIORITY_URGENT    = 0x0800
        PRIORITY_LOW       = 0x0C00

        # Bit 1/1
        NO_ACK_REQUESTED = 0x0000
        ACK_REQUESTED    = 0x0200

        #Bit 1/0
        CONFIRM_NO_ERROR = 0x0000
        CONFIRM_ERROR    = 0x0100

        # Bit 0/7
        DESTINATION_INDIVIDUAL_ADDRESS = 0x0000
        DESTINATION_GROUP_ADDRESS      = 0x0080

        # Bit 0/6+5+4
        HOP_COUNT_NO = 0x0070
        HOP_COUNT_1ST = 0x0060

        # Bit 0/3+2+1+0
        STANDARD_FRAME_FORMAT = 0x0000
        EXTENDED_FRAME_FORMAT = 0x0001





""" Implementation of Basic KNX Floats """

""" See: http://www.knx.org/fileadmin/template/documents/downloads_support_menu/KNX_tutor_seminar_page/Advanced_documentation/05_Interworking_E1209.pdf as reference """

from enum import Enum
import time
from .dpt import DPT_Base,ConversionError

class DPT_Float(DPT_Base):
    """ Abstraction for KNX 2 Octet Floating Point Numbers """
    """ DPT 9.xxx """

    value_min = -671088.64
    value_max = 670760.96
    unit = ""
    resolution = 1

    @classmethod
    def from_knx(cls, raw):
        """Convert a 2 byte KNX float to a flaot value"""
        cls.test_bytesarray(raw,2)

        data = (raw[0] * 256 ) + raw[1]
        exponent = (data >> 11) & 0x0f
        significand = data & 0x7ff
        sign = data >> 15

        if sign == 1:
            significand = significand - 2048

        value =  float(significand << exponent) / 100

        if not cls.test_boundaries(value):
            raise ConversionError(value)

        return value

    @classmethod
    def to_knx(cls, value):
        """Convert a float to a 2 byte KNX float value"""

        if not isinstance(value, (int, float)):
            raise ConversionError(value)

        if not cls.test_boundaries(value):
            raise ConversionError(value)

        sign = 1 if value < 0 else 0

        def calc_exponent(value, sign):
            exponent = 0
            significand = abs ( int (value * 100) )

            while significand < -2048 or significand > 2048:
                exponent += 1
                significand >>= 1

            if sign:
                significand ^= 0x7ff # invert
                significand += 1     # and add 1

            return exponent,significand

        exponent,significand = calc_exponent(value, sign)

        return (sign << 7) | (exponent << 3) | (significand >> 8), significand & 0xff

    @classmethod
    def test_boundaries(cls, value):
        return value >= cls.value_min and \
               value <= cls.value_max

class DPT_Temperature(DPT_Float):
    """ Abstraction for KNX 2 Octet Floating Point Numbers """
    """ DPT 9.001 """

    value_min = -273
    value_max = 670760
    unit = "C"
    resolution = 1

class DPT_Lux(DPT_Float):
    """ Abstraction for KNX 2 Octet Floating Point Numbers """
    """ DPT 9.004 """

    value_min = 0
    value_max = 670760
    unit = "Lux"
    resolution = 1

class DPT_Humidity(DPT_Float):
    """ Abstraction for KNX 2 Octet Floating Point Numbers """
    """ DPT 9.007 """

    value_min = 0
    value_max = 670760
    unit = "%"
    resolution = 1




import threading
import time

class StateUpdater(threading.Thread):

    def __init__(self, xknx, timeout, start_timeout=15):
        self.xknx = xknx
        self.timeout = timeout
        self.start_timeout = start_timeout
        threading.Thread.__init__(self)


    def run(self):
        time.sleep(self.start_timeout)
        print("Starting Update Thread")
        while True:
            self.request_states()
            time.sleep(self.timeout)

    def request_states(self):
        for device in self.xknx.devices.devices:
            device.request_state()

    @staticmethod
    def start_thread(xknx, timeout = 60):
        t = StateUpdater(xknx, timeout)
        t.setDaemon(True)
        t.start()





class Device:
    def __init__(self, xknx, name):
        self.xknx = xknx
        self.after_update_callback = lambda x: None
        self.name = name

    def request_state(self):
        pass

    def process(self,telegram):
        pass

    def get_name(self):
        return self.name

    def do(self,action):
        pass



import time
import threading

class CouldNotResolveAddress(Exception):

    def __init__(self, group_address):
        self.group_address = group_address

    def __str__(self):
        return "CouldNotResolveAddress <name={0}>".format(self.group_address)

class CouldNotResolveName(Exception):

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "CouldNotResolveName <name={0}>".format(self.name)


class Devices:

    def __init__(self):
        self.devices = []

    def device_by_group_address( self, group_address):
        for device in self.devices:
            if device.has_group_address(group_address):
                return device
        raise CouldNotResolveAddress(group_address)

    def device_by_name( self, name):
        for device in self.devices:
            if device.name == name:
                return device
        raise CouldNotResolveName(name)

    def get_devices( self ):
        return self.devices




from .device import Device
from .address import Address
from .dpt_float import DPT_Temperature
from .exception import CouldNotParseTelegram
from .dpt import DPT_Array

import time


class Thermostat(Device):
    def __init__(self, xknx, name, config):
        Device.__init__(self, xknx, name)
        self.group_address = Address(config["group_address"])
        self.last_set = time.time();
        self.temperature = 0

    def has_group_address(self, group_address):
        return self.group_address == group_address

    def process(self,telegram):
        if not isinstance(telegram.payload, DPT_Array) or len(telegram.payload.value) != 2:
            raise CouldNotParseTelegram()

        self.temperature = DPT_Temperature().from_knx(( telegram.payload.value[0] , telegram.payload.value[1] ))

        self.after_update_callback(self)

    def __str__(self):
        return "<Thermostat group_address={0}, name={1}>".format(self.group_address,self.name)



from setuptools import setup, find_packages

setup(
        name='xknx',
        description='A Wrapper around KNX protocol.',

        version='0.3',
        download_url='https://github.com/XKNX/xknx/tarball/0.3',
        url='http://xknx.io/',

        author='Julius Mittenzwei',
        author_email='julius@mittenzwei.com',
        license='MIT',
        classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Developers',
          'Topic :: System :: Hardware :: Hardware Drivers',
          'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5'
        ],
        packages=find_packages(),
        install_requires=['PyYAML'],
        keywords = 'knx ip knxip home automation',
        zip_safe=False)



from .address import Address
from .telegram import Telegram,TelegramType
from .device import Device
from .address import Address
from .travelcalculator import TravelCalculator
from .exception import CouldNotParseTelegram
from .dpt import DPT_Binary,DPT_Array

class Shutter(Device):

    def __init__(self, xknx, name, config):
        Device.__init__(self, xknx, name)
        self.group_address_long = Address(config.get("group_address_long"))
        self.group_address_short = Address(config.get("group_address_short"))
        self.group_address_position = Address(config.get("group_address_position"))
        self.group_address_position_feedback = Address(config.get("group_address_position_feedback"))

        # Assuming 20 seconds is the typical travelling time of an average cover ...
        travelling_time_down = config.get("travelling_time_down", 20)
        travelling_time_up = config.get("travelling_time_up", 22)

        self.travelcalculator = TravelCalculator(travelling_time_down,travelling_time_up)


    def has_group_address(self, group_address):
        return ( self.group_address_long == group_address ) or (self.group_address_short == group_address ) or (self.group_address_position_feedback == group_address )


    def supports_direct_positioning(self):
        return self.group_address_position.is_set()


    def __str__(self):
        return "<Shutter group_address_long={0}, group_address_short={1}, group_address_position, group_address_position_feedback={2}, name={3}>".format(self.group_address_long,self.group_address_short,self.group_address_position, self.group_address_position_feedback,self.name)


    def send(self, group_address, payload):
        telegram = Telegram()
        telegram.group_address=group_address
        telegram.payload = payload
        self.xknx.telegrams.put(telegram)


    def set_down(self):
        if not self.group_address_long.is_set():
            print("group_address_long not defined for device {0}".format(self.get_name()))
            return
        self.send(self.group_address_long, DPT_Binary(1) )
        self.travelcalculator.start_travel_down()


    def set_up(self):
        if not self.group_address_long.is_set():
            print("group_address_long not defined for device {0}".format(self.get_name()))
            return
        self.send(self.group_address_long, DPT_Binary(0) )
        self.travelcalculator.start_travel_up()


    def set_short_down(self):
        if not self.group_address_short.is_set():
            print("group_address_short not defined for device {0}".format(self.get_name()))
            return
        self.send(self.group_address_short, DPT_Binary(1) )


    def set_short_up(self):
        if not self.group_address_short.is_set():
            print("group_address_short not defined for device {0}".format(self.get_name()))
            return
        self.send(self.group_address_short, DPT_Binary(0) )


    def stop(self):
        # Thats the KNX way of doing this. electrical engineers ... m-)
        self.set_short_down()
        self.travelcalculator.stop()


    def set_position(self, position):
        if not self.supports_direct_positioning():

            current_position = self.current_position()
            if position > current_position:
                self.send(self.group_address_long, DPT_Binary(1) )
            elif position < current_position:
                self.send(self.group_address_long, DPT_Binary(0) )
            self.travelcalculator.start_travel( position )
            return
        self.send(self.group_address_position, DPT_Array(position))
        self.travelcalculator.start_travel( position )


    def auto_stop_if_necessary(self):
        # If device does not support auto_positioning,
        # we have to ttop the device when position is reached.
        # unless device was travelling to fully open
        # or fully closed state
        if (
            not self.supports_direct_positioning() and
                self.position_reached() and
                not self.is_open() and
                not self.is_closed()
            ):
            self.stop()


    def do(self,action):
        if(action=="up"):
            self.set_up()
        elif(action=="short_up"):
            self.set_short_up()
        elif(action=="down"):
            self.set_down()
        elif(action=="short_down"):
            self.set_short_down()        
        else:
            print("{0}: Could not understand action {1}".format(self.get_name(), action))


    def request_state(self):
        if not self.group_address_position_feedback.is_set():
            print("group_position not defined for device {0}".format(self.get_name()))
            return
        if self.travelcalculator.is_travelling():
            # Cover is travelling, requesting state will return false results
            return

        telegram = Telegram(self.group_address_position_feedback, TelegramType.GROUP_READ)
        self.xknx.telegrams.put(telegram)


    def process(self,telegram):
        if not isinstance(telegram.payload, DPT_Array) or len(telegram.payload.value) != 1:
            raise CouldNotParseTelegram()

        self.travelcalculator.set_position( telegram.payload.value[0] )
        self.after_update_callback(self)


    def current_position(self):
        return self.travelcalculator.current_position()


    def is_travelling(self):
        return self.travelcalculator.is_travelling()


    def position_reached(self):
        return self.travelcalculator.position_reached()


    def is_open(self):
        return self.travelcalculator.is_open()


    def is_closed(self):
        return self.travelcalculator.is_closed()




from .device import Device
from .telegram import Telegram,TelegramType
from .address import Address
from .exception import CouldNotParseTelegram
from .dpt import DPT_Binary,DPT_Array
import time

class Dimmer(Device):
    def __init__(self, xknx, name, config):
        Device.__init__(self, xknx, name)
        self.group_address_switch = Address(config.get("group_address_switch"))
        self.group_address_dimm = Address(config.get("group_address_dimm"))
        self.group_address_dimm_feedback = Address(config.get("group_address_dimm_feedback"))
        self.group_address_state = Address(config.get("group_address_state"))
 
        self.state = False
        self.brightness = 0

    def has_group_address(self, group_address):
        return ( self.group_address_state == group_address ) or ( self.group_address_switch == group_address ) or (self.group_address_dimm == group_address ) or (self.group_address_dimm_feedback == group_address )

    def __str__(self):
        return "<Dimmer group_address_switch={0}, group_address_dimm={0}, group_address_dimm_feedback={2},name={3}>".format(self.group_address_switch,self.group_address_dimm,self.group_address_dimm_feedback,self.name)

    def set_internal_state(self, state):
        if state != self.state:
            self.state = state
            self.after_update_callback(self)

    def set_internal_brightness(self, brightness):
        if brightness != self.brightness:
            self.brightness = brightness
            self.after_update_callback(self)


    def send(self, group_address, payload = None):
        telegram = Telegram()
        telegram.group_address=group_address

        telegram.payload = payload
        self.xknx.telegrams.put(telegram)

    def set_on(self):
        self.send(self.group_address_switch, DPT_Binary(1) )
        self.set_internal_state(True)

    def set_off(self):
        self.send(self.group_address_switch, DPT_Binary(0) )
        self.set_internal_state(False)

    def set_brightness(self, brightness):
        self.send(self.group_address_dimm_feedback, DPT_Array(brightness) )
        self.set_internal_brightness(brightness)


    def request_state(self):
        if not self.group_address_dimm_feedback.is_set():
            print("group_address_dimm_feedback not defined for device {0}".format(self.get_name()))
            return

        telegram_switch = Telegram(self.group_address_switch, TelegramType.GROUP_READ)
        self.xknx.telegrams.put(telegram_switch)
        telegram_dimm = Telegram(self.group_address_dimm_feedback, TelegramType.GROUP_READ)
        self.xknx.telegrams.put(telegram_dimm)


    def process(self,telegram):
        if telegram.group_address == self.group_address_switch or (telegram.group_address == self.group_address_state):
            self._process_state(telegram)
        elif telegram.group_address == self.group_address_dimm_feedback:
            self._process_dimm(telegram) 

    def _process_dimm(self,telegram):
        if not isinstance(telegram.payload, DPT_Array) or len(telegram.payload.value) != 1:
            raise CouldNotParseTelegram()

        self.set_internal_brightness(telegram.payload.value[0])

    def _process_state(self,telegram):

        if not isinstance( telegram.payload, DPT_Binary ):
            raise CouldNotParseTelegram()

        if telegram.payload.value == 0 :
            self.set_internal_state(False)
        elif telegram.payload.value == 1 :
            self.set_internal_state(True)
        else:
            raise CouldNotParseTelegram()




from .xknx import XKNX
from .address import Address,CouldNotParseAddress,AddressType,AddressFormat
from .telegram import Telegram,TelegramDirection,TelegramType
from .knxip import KNXIPFrame,CEMIFrame
from .knxip_enum import KNXIPServiceType,CEMIMessageCode,APCICommand,CEMIFlags
from .stateupdater import StateUpdater
from .multicast import Multicast,MulticastDaemon
from .telegram_processor import TelegramProcessor
from .devices import Devices,CouldNotResolveAddress
from .binaryinput import BinaryInput,BinaryInputState
from .switch import Switch,Action
from .shutter import Shutter
from .travelcalculator import TravelCalculator
from .thermostat import Thermostat
from .dimmer import Dimmer
from .outlet import Outlet
from .time import Time
from .config import Config
from .dpt import DPT_Base,DPT_Binary,DPT_Array,ConversionError,DPT_Comparator
from .dpt_float import DPT_Float,DPT_Lux,DPT_Temperature,DPT_Humidity
from .dpt_time import DPT_Time,DPT_Weekday
from .globals import Globals
